import { IByQueryArg, IndexedSelectorOrArr, IQueryArg } from 'gs-dom';
import { RequiredSome } from 'gs-base';

type FillQueryArg = Omit<IByQueryArg, 'gdom'>;
type TopQuery = IndexedSelectorOrArr | FillQueryArg;
type LowerQuery = IndexedSelectorOrArr | IQueryArg;
interface IFieldQueries {
    label?: LowerQuery;
    field?: LowerQuery;
}
interface IChildQueries extends IFieldQueries {
    title?: LowerQuery;
    child?: LowerQuery;
    childId?: LowerQuery;
    childEdit?: LowerQuery;
}
interface IGroupQueries extends IChildQueries {
    group?: LowerQuery;
    groupAdd?: LowerQuery;
}
interface IRootQueries extends IGroupQueries {
    root?: TopQuery;
}

interface IDataProvider {
    readonly global: Object;
    readonly children: Record<string, IChildDataProvider>;
    readonly groups: Record<string, IGroupDataProvider>;
}
interface IChildDataProvider {
    readonly title?: string;
    readonly data: Object;
}
interface IGroupId {
    readonly field?: string;
    readonly pattern?: string;
}
interface IGroupDataProvider {
    readonly title?: string;
    readonly id?: IGroupId;
    readonly data: Object[];
}

interface IActionInject<Inject extends string = 'inject' | 'react'> {
    inject: Inject;
    delay?: number;
}
interface IAction<Type extends string, Inject extends string> {
    type: Type;
    /**
     * 是否注入操作
     * - 为数值时,代表启用的延迟
     */
    inject?: number | Inject | IActionInject<Inject>;
    /**
     * 操作结束后的延迟
     */
    afterDelay?: number;
}
interface ILabeled {
    /**
     * 用于检测字段名称的正则字符串
     */
    label?: string | RegExp;
}
interface IActionQuery extends ILabeled {
    /**
     * 用于检测字段`HTML`元素的查询器
     */
    query?: LowerQuery;
    /**
     * 操作该元素时,是否滚动到该元素
     */
    scroll?: boolean | ScrollIntoViewOptions;
}

interface ICmd<T extends string = string, Inject extends string = 'inject' | 'react'> extends IAction<T, Inject> {
}
interface IFnCmd extends ICmd<'fn'> {
    fnName: string;
    args?: any;
}
interface IElCmd<T extends string = string> extends ICmd<T>, IActionQuery {
}
type ELCmd<T extends string> = RequiredSome<IElCmd<T>, 'query'> | RequiredSome<IElCmd<T>, 'label'>;
type ClickCmd = ELCmd<'click'>;

/**
 * 值映射器,用于将原值进行替换
 * - `value数组`的第一个值为`正则字符串`或`正则`
 * - 若值满足`正则字符串`条件，则将映射为`value数组`的第二个值
 */
interface IValueMap {
    value: [string | RegExp, string | number | boolean][];
    default?: string | number | boolean;
}
interface IRegexReplace {
    pattern: string | RegExp;
    /**
     * 根据正则进行格式化
     * - `$0` 表示原值,若存在`map`,则映射为`map[0]`的值
     * - `$1` 表示第一个子数组,若存在`map`,则映射为`map[1]`的值,以此类推
     */
    fmt?: string;
    map?: IValueMap[];
}
/**
 * 用于将原值进行替换的正则字符串
 * - 若`replace`为正则字符串,则将满足条件的皆替换为空
 * - 若`replace`为`ValueMap`,则将映射为对应值
 */
type RegexReplace = string | RegExp | IValueMap | IRegexReplace;

interface IValueCheck {
    /**
     * 当查询不为空时,为存在值
     */
    query: LowerQuery;
    /**
     * 查询到元素时,用于检测值的属性,
     * - 当设置了`nonValue`时,默认值为 `innerText` 与 `value`
     * - 当没有设置`nonValue`时,将仅检测元素是否存在
     */
    attr?: string;
    /**
     * 正则字符串
     * > 当节点值不匹配时,代表存在值
     */
    nonValue?: string | RegExp;
}

type SetValue = string | number | boolean;
type MatchValue = SetValue | RegExp;
type SetValueMap = MatchValue | IValueMap;
type ObjectKey = string | number | symbol;
/**
 * 设置器值查找顺序为
 *  1.`value`
 *  2.`cxt`
 *  3.`global`
 */
interface ISetter<CK extends ObjectKey = ObjectKey, GK extends ObjectKey = ObjectKey, Type extends string = string, Parent extends string | undefined = string, Inject extends string = 'react'> extends ISetterBase, IAction<Type, Inject>, IActionQuery {
    /**
     * 在当前上下文环境查找的字段`key`,可能为`group`,`children`,也可能为`global`
     */
    cxt?: CK;
    /**
     * 一定在`global`中查找数据
     */
    global?: GK;
    /**
     * 用于将原值进行替换的正则字符串
     * - 若`replace`为正则字符串,则将满足条件的皆替换为空
     * - 若`replace`为`ValueMap`,则将映射为对应值
     */
    replace?: RegexReplace | RegexReplace[];
    /**
     * 存在值时是否覆盖
     */
    overwrite?: boolean;
    /**
     * 检查值存在设置
     */
    valueCheck?: IValueCheck | LowerQuery;
    /**
     * 父级配置
     */
    parent?: Parent;
}
interface ISetterBase {
    /**
     * 用于设置器的值
     */
    value?: SetValue | SetValue[];
    /**
     * 用于匹配的值
     * - 若不存在则根据`value`或`defaultValue`数据生成
     * - 若为`string`,则将其生成为正则
     * - 若为`MatchMap`，则将`value`映射为对应值
     */
    matchValue?: SetValueMap | SetValueMap[];
    /**
     * 用于匹配的值时的默认匹配,非正则将会被转换为正则
     */
    matchDefault?: MatchValue | MatchValue[];
}
type AnySetter = ISetter;

interface ISetTextQueries {
    inputQuery?: LowerQuery;
}
interface ISetText<CK extends ObjectKey, GK extends ObjectKey, Parent extends string | undefined = string> extends Omit<ISetter<CK, GK, 'text', Parent>, 'matchValue'>, ISetTextQueries {
}

interface ISetDropdownBase extends ISetTextQueries {
    pull?: boolean;
    /**
     * 在`pull`或`search`时,输入值的长度
     */
    valueLength?: number;
    /**
     * 每一步动作的延迟
     */
    actionDelay?: number;
    menuQuery?: LowerQuery;
    confirmQuery?: LowerQuery;
    cancelBtnQuery?: LowerQuery;
    /**
     * 等待菜单的延迟
     */
    waitMenuDelay?: number;
}
interface ISetDropdown<CK extends ObjectKey, GK extends ObjectKey, Parent extends string | undefined = string> extends ISetter<CK, GK, 'dropdown', Parent>, ISetDropdownBase {
    /**
     * 等待搜索的延迟
     */
    waitSearchDelay?: number;
    itemQuery?: LowerQuery;
    /**
     * 用于多级值时,获取其`父级值HTML元素`的查询器
     */
    itemParentQuery?: LowerQuery;
    itemClickQuery?: LowerQuery;
    searchQuery?: LowerQuery;
    searchBtnQuery?: LowerQuery;
}

interface ISetDate<CK extends ObjectKey, GK extends ObjectKey, Parent extends string | undefined = string> extends ISetter<CK, GK, 'date', Parent>, ISetDropdownBase {
    yearValueQuery?: LowerQuery;
    yearPrevQuery?: LowerQuery;
    yearNextQuery?: LowerQuery;
    monthItemQuery?: LowerQuery;
    monthValueQuery?: LowerQuery;
    monthNextQuery?: LowerQuery;
    monthPrevQuery?: LowerQuery;
    dayItemQuery?: LowerQuery;
}

interface IField<ST extends string> extends ILabeled, ISetterBase {
    /**
     * 设置器类型,默认为`text`
     * - `text` 为文本设置器
     * - `dropdown` 为下拉菜单设置器
     */
    setter?: ST;
}

/**
 * 各种`命令`与`设置器`的默认配置名
 */
interface IActionOptions {
    text?: string;
    date?: string;
    dropdown?: string;
    radio?: string;
    step?: string;
    /**
     * 常规操作后延迟
     */
    actionDelay?: number;
    /**
     * 弹出菜单操作后延迟
     */
    dropdownDelay?: number;
}

type PageAction<CxtKey extends ObjectKey, GlobalKey extends ObjectKey> = CxtKey | ICmd | ISetter<CxtKey, GlobalKey>;
interface IPageRule<Data extends IDataProvider> {
    /**
     * 用于匹配`网站URL`的正则
     */
    url: string | RegExp;
    queries?: IRootQueries;
    /**
     * 动作选项配置
     */
    options?: IActionOptions;
    before?: PageAction<keyof Data['global'], keyof Data['global']>[];
    child?: IChildRule<Data>[];
    group?: IGroupRule<Data>[];
    after?: PageAction<keyof Data['global'], keyof Data['global']>[];
}
interface IChildRule<Data extends IDataProvider, Child extends keyof Data['children'] = keyof Data['children']> {
    childKey: Child;
    title: string | RegExp;
    queries?: IChildQueries;
    options?: IActionOptions;
    childRules: PageAction<keyof Data['children'][Child]['data'], keyof Data['global']>[];
}
interface IGroupRule<Data extends IDataProvider, Group extends keyof Data['groups'] = keyof Data['groups']> {
    groupKey: Group;
    title: string | RegExp;
    queries?: IGroupQueries;
    options?: IActionOptions;
    groupRules: PageAction<keyof Data['groups'][Group]['data'][0], keyof Data['global']>[];
    condRules?: ICondRule<keyof Data['groups'][Group]['data'][0], keyof Data['global']>[];
}
interface ICondRule<CxtKey extends ObjectKey, GlobalKey extends ObjectKey> {
    cxtKey: CxtKey;
    cond: string | RegExp;
    trueRules: PageAction<CxtKey, GlobalKey>[];
    falseRules?: PageAction<CxtKey, GlobalKey>[];
}

declare function toCfgJson(data: any, space?: number): string;

interface ISetRadio<CK extends ObjectKey, GK extends ObjectKey, Parent extends string | undefined = string> extends ISetter<CK, GK, 'radio', Parent> {
    radioQuery?: LowerQuery;
    wrapperQuery?: LowerQuery;
}

interface ISetStep<CK extends ObjectKey, GK extends ObjectKey, Parent extends string | undefined = string> extends ISetter<CK, GK, 'step', Parent>, ISetDropdownBase {
    itemQueries?: LowerQuery[];
    stepDelay?: number;
    firstValue?: string;
    useDom?: boolean;
}

type NonType<A extends IAction<any, any>> = Omit<A, 'type'>;
type SetArg<Setter extends ISetter> = NonType<RequiredSome<Setter, 'cxt'>> | NonType<RequiredSome<Setter, 'global'>>;
declare class ActionBuilder<Page, CxtKey extends ObjectKey, GlobalKey extends ObjectKey, Options extends IActionOptions, Ab extends typeof this = typeof this> {
    protected actions: PageAction<CxtKey, GlobalKey>[];
    protected page: Page;
    constructor(page: Page, actions: PageAction<CxtKey, GlobalKey>[]);
    get end(): Page;
    date(...args: (CxtKey | SetArg<ISetDate<CxtKey, GlobalKey, Options['date']>>)[]): Ab;
    dropdown(...args: (CxtKey | SetArg<ISetDropdown<CxtKey, GlobalKey, Options['dropdown']>>)[]): Ab;
    radio(...args: (CxtKey | SetArg<ISetRadio<CxtKey, GlobalKey, Options['radio']>>)[]): Ab;
    step(...args: (CxtKey | SetArg<ISetStep<CxtKey, GlobalKey, Options['step']>>)[]): Ab;
    text(...args: (CxtKey | SetArg<ISetText<CxtKey, GlobalKey, Options['text']>>)[]): Ab;
    click(...args: NonType<ClickCmd>[]): Ab;
    fn(...args: NonType<IFnCmd>[]): Ab;
    key(...args: CxtKey[]): Ab;
    protected action(type: string, args: string | IAction<any, any>[]): Ab;
}

interface IBuildActionOptions extends IActionOptions {
}
type BuildCmd = ClickCmd | IFnCmd;
type BuildAction<CxtKey extends ObjectKey, GlobalKey extends ObjectKey, Options extends IActionOptions = IBuildActionOptions> = CxtKey | BuildCmd | ISetDate<CxtKey, GlobalKey, Options['date']> | ISetDropdown<CxtKey, GlobalKey, Options['dropdown']> | ISetRadio<CxtKey, GlobalKey, Options['radio']> | ISetStep<CxtKey, GlobalKey, Options['step']> | ISetText<CxtKey, GlobalKey, Options['text']>;

declare class CondFalseBuilder<Page extends PageRuleBuilder<Data, Options>, Parent, CxtKey extends ObjectKey, Data extends IDataProvider, Options extends IActionOptions> extends ActionBuilder<Page, CxtKey, keyof Data['global'], Options, typeof this> {
    protected parent: Parent;
    constructor(page: Page, parent: Parent, actions: PageAction<CxtKey, keyof Data["global"]>[]);
    get back(): Parent;
}
declare class CondBuilder<Page extends PageRuleBuilder<Data, Options>, Parent, CxtKey extends ObjectKey, Data extends IDataProvider, Options extends IActionOptions> extends CondFalseBuilder<Page, Parent, CxtKey, Data, Options> {
    protected condRule: ICondRule<CxtKey, keyof Data['global']>;
    constructor(page: Page, parent: Parent, condRules: ICondRule<any, keyof Data["global"]>[], cxtKey: CxtKey, cond: string | RegExp);
    addFalseRules(): CondFalseBuilder<Page, Parent, CxtKey, Data, Options>;
}

declare class GroupBuilder<Page extends PageRuleBuilder<Data, Options>, Data extends IDataProvider, Group extends keyof Data['groups'], Options extends IActionOptions> extends ActionBuilder<Page, keyof Data['groups'][Group]['data'][0], keyof Data['global'], Options, typeof this> {
    protected groupRule: IGroupRule<Data, Group>;
    constructor(page: Page, groups: IGroupRule<Data, any>[], groupKey: Group, title: string | RegExp);
    query(queries: IRootQueries): this;
    options(options: Options): this;
    addCondRules(cxtKey: keyof Data['groups'][Group]['data'][0], cond: string | RegExp): CondBuilder<Page, this, keyof Data["groups"][Group]["data"][0], Data, Options>;
}

declare function buildPage<Data extends IDataProvider, Options extends IActionOptions = IBuildActionOptions>(url: string, queries?: IRootQueries): PageRuleBuilder<Data, Options>;
declare class PageRuleBuilder<Data extends IDataProvider, Options extends IActionOptions> {
    #private;
    constructor(url: string, queries?: IRootQueries);
    query(queries: IRootQueries): this;
    options(options: Options): this;
    addBefore(): ActionBuilder<this, keyof Data["global"], keyof Data["global"], Options>;
    addChild<Child extends keyof Data['children'] = keyof Data['children']>(childKey: keyof Data['children'], title: string | RegExp): ChildBuilder<this, Data, Child, Options>;
    addGroup<Group extends keyof Data['groups'] = keyof Data['groups']>(groupKey: Group, title: string | RegExp): GroupBuilder<this, Data, Group, Options>;
    addAfter(): ActionBuilder<this, keyof Data["global"], keyof Data["global"], Options>;
    toJSON(): IPageRule<Data>;
    build(): IPageRule<Data>;
    toString(): string;
}

declare class ChildBuilder<Page extends PageRuleBuilder<Data, Options>, Data extends IDataProvider, Child extends keyof Data['children'], Options extends IActionOptions> extends ActionBuilder<Page, keyof Data['children'][Child]['data'], keyof Data['global'], Options, typeof this> {
    protected childRule: IChildRule<Data, Child>;
    constructor(page: Page, children: IChildRule<Data, any>[], childKey: Child, title: string | RegExp);
    query(queries: IRootQueries): this;
    options(options: Options): this;
}

declare const DefaultSetter: string;
declare function setSetterOption<Cfg extends AnySetter = AnySetter>(parent: string, cfg: Cfg): void;
declare function getSetterOption<Cfg extends AnySetter = AnySetter>(cfg: Cfg, parent?: string): Cfg;
declare function exportSetterOptions(space?: number): string;
declare function importSetterOptions(json: string): void;
declare function getSetterOptionType(wrap?: boolean): Record<string, string[]> | string;

export { ActionBuilder, type AnySetter, type BuildAction, type BuildCmd, ChildBuilder, type ClickCmd, CondBuilder, CondFalseBuilder, DefaultSetter, type ELCmd, type FillQueryArg, GroupBuilder, type IAction, type IActionInject, type IActionQuery, type IBuildActionOptions, type IChildDataProvider, type IChildQueries, type IChildRule, type ICmd, type ICondRule, type IDataProvider, type IElCmd, type IField, type IFieldQueries, type IFnCmd, type IGroupDataProvider, type IGroupId, type IGroupQueries, type IGroupRule, type ILabeled, type IPageRule, type IRegexReplace, type IRootQueries, type ISetDate, type ISetDropdown, type ISetDropdownBase, type ISetText, type ISetTextQueries, type ISetter, type ISetterBase, type IValueCheck, type IValueMap, type LowerQuery, type MatchValue, type NonType, type ObjectKey, type PageAction, PageRuleBuilder, type RegexReplace, type SetArg, type SetValue, type SetValueMap, type TopQuery, buildPage, exportSetterOptions, getSetterOption, getSetterOptionType, importSetterOptions, setSetterOption, toCfgJson };
